package global

var (
	Version               = "dev"
	Commit                = ""
	BuildTime             = "none"
	Debug                 = false
	ProfilingEnabled      = false
	ConfigFilename        = "cog.yaml"
	ReplicateRegistryHost = "r8.im"
	ReplicateWebsiteHost  = "replicate.com"
	LabelNamespace        = "run.cog."
)
